==> playship-carroms-1.1.5.tgz: package/src/logic/wheeloflife.js <==
const oddUtils = require('../utils/oddUtils');
const logger = oddUtils.getLogger();
const turnFillUtils = require('../utils/fillAndFetchUtils').turnFillUtils;
const genericFetchUtils = require(
    '../utils/fillAndFetchUtils').genericFetchUtils;
const TurnAndReconcileStateHolder = require('../utils/turnAndReconHolder');

const ALL_PROP_KEYS = oddUtils.getAllPropKeys();
const ALL_PROP_VALUES = oddUtils.getALLPropValues();
const ALL_TRANSIENT_KEYS = oddUtils.getAllTransientKeys();

const nextTurn = function(game, isFoul) {
  turnInit(game, parseInt(game.getCurrentTurnIndex()) + 1);
  game.nextTurn(isFoul);
  // issue save point here
  issueTurnStartSavePoint(game);
};

const continueTurn = function(game, isFoul) {
  tackInit(game, parseInt(game.getCurrentTurnIndex()));
  game.continueTurn(isFoul);
  // issue save point here
  issueTurnStartSavePoint(game);
};

function issueTurnStartSavePoint(game) {
  const gameKey = game.getId();
  game.issueSavepoint('__turn_start', err => {
    if (err) return logger.error({err: err, key: gameKey},
        'Error in issuing savepoint');
    logger.debug({key: gameKey}, 'pre-turn save point issued');
  });
}

function turnInit(game, turnIndex) {
  game.setGameProp(ALL_PROP_KEYS.POTT_TACK, 0);
  // initialize empty turn cache for fill fetch utils
  turnFillUtils.initEmptyTurnCache(game, turnIndex | 0);
  tackInit(game, turnIndex | 0);
}

function tackInit(game, currentTurnIndex) {
  let currentTack = parseInt(game.getGameProp(ALL_PROP_KEYS.POTT_TACK));
  // initialize empty tack cache
  turnFillUtils.initEmptyTackCache(game, currentTurnIndex,
      currentTack + 1);
  // increment and set current tack
  game.setGameProp(ALL_PROP_KEYS.POTT_TACK, currentTack + 1);
  let striker = genericFetchUtils.fetchStriker(game);
  striker.setProp(ALL_PROP_KEYS.IS_POTTED, false);

  game.setGameProp(ALL_PROP_KEYS.RECONCILE_STATUS, null);
  game.removeTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER);

  for (let p of game.getAllPlayerIds()) {
    game.setPlayerProp(p, ALL_PROP_KEYS.GRACE_TIME_START, null);
    game.setPlayerProp(p, ALL_PROP_KEYS.IS_FORCE_TO_BE_REPLAYED, false);
  }

  game.setGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT, false);
}

const grantGraceTimeForTurn = function(game, playerId) {
  let prevGraceTimeLeftForTurn = parseInt(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.GRACE_TIME_LEFT_FOR_TURN));
  let totalGraceTimeLeft = parseInt(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT));

  // bucketing the grace time, for each turn
  let graceTimeToGrantForCurTurn = Math.max(
      ALL_PROP_VALUES.MIN_GRACE_TIME_TO_GRANT,
      Math.min(totalGraceTimeLeft, ALL_PROP_VALUES.GRACE_TIME_BUCKET_FOR_TURN));

  game.setPlayerProp(playerId, ALL_PROP_KEYS.GRACE_TIME_LEFT_FOR_TURN,
      graceTimeToGrantForCurTurn);
  return prevGraceTimeLeftForTurn !== graceTimeToGrantForCurTurn;
};

const startGame = function(game, cb) {
  turnInit(game);
  issueTurnStartSavePoint(game);
  game.initCollisionMap(cb);
};

const playerJoined = function(game, playerId) {
  const gameKey = game.getId();
  let oppId = genericFetchUtils.fetchOpponentId(game, playerId);
  let playerColor = game.getPlayerProp(playerId, ALL_PROP_KEYS.COLOR);
  let oppColor = game.getPlayerProp(oppId, ALL_PROP_KEYS.COLOR);
  if (!playerColor || !oppColor) {
    // if testing always first player is white, black
    if (process.env.TESTING) {
      playerColor = ALL_PROP_VALUES.COLOR.WHITE;
      oppColor = ALL_PROP_VALUES.COLOR.BLACK;
    } else {
      playerColor = Math.random() * 2 | 0 ?
          ALL_PROP_VALUES.COLOR.WHITE :
          ALL_PROP_VALUES.COLOR.BLACK;
      oppColor = playerColor === ALL_PROP_VALUES.COLOR.WHITE ?
          ALL_PROP_VALUES.COLOR.BLACK :
          ALL_PROP_VALUES.COLOR.WHITE;
    }
    logger.debug({key: gameKey}, 'Setting color', playerColor, 'for', playerId,
        'and', oppColor, 'for', oppId);
    game.setPlayerProp(playerId, ALL_PROP_KEYS.COLOR, playerColor);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.COLOR, oppColor);
    // for 1 player set rotation true
    game.setPlayerProp(playerId, ALL_PROP_KEYS.IS_ROTATION, false);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.IS_ROTATION, true);

    game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_POCKETED_RED, false);
    game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_SCORED_RED, false);

    game.setPlayerProp(oppId, ALL_PROP_KEYS.DID_POCKETED_RED, false);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.DID_SCORED_RED, false);

    game.setPlayerProp(playerId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        ALL_PROP_VALUES.GRACE_TIME_GRANTED);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        ALL_PROP_VALUES.GRACE_TIME_GRANTED);

    grantGraceTimeForTurn(game, playerId);
    grantGraceTimeForTurn(game, oppId);
  }

  // if player has the turn and connected back when force needs to be replayed
  // just revert to previous save point
  let curPlId = game.getCurrentTurnPlayerId();
  if (curPlId === playerId && oddUtils.parseBoolean(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.IS_FORCE_TO_BE_REPLAYED))) {
    logger.info({key: gameKey}, 'Force needs to be replyaed fo the game');
    revertToPreviousSavePoint(game);
    return game.getState() >= 30; // telling neutrino to not send game state, if game
  }
  // based on should send game, default value is true
  return oddUtils.parseBoolean(
      game.getGameProp(ALL_PROP_KEYS.SHOULD_SEND_GAME_STATE), true);
};

const shouldGameBeEnded = function(game, judgment, playerId) {
  const gameKey = game.getId();
  let allPlayerIds = game.getAllPlayerIds();
  if (allPlayerIds.length !== 2) throw new Error(
      'Number of playerIds available is not 2. ' + allPlayerIds.length);
  let playerIdA = allPlayerIds[0];
  let playerIdB = allPlayerIds[1];

  let playerAScoreDelta, playerBScoreDelta;
  if (playerId === playerIdA) {
    playerAScoreDelta = judgment.selfScore;
    playerBScoreDelta = judgment.oppScore;
  } else {
    playerBScoreDelta = judgment.selfScore;
    playerAScoreDelta = judgment.oppScore;
  }

  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let playerAScore = game.getPlayerScore(playerIdA) + playerAScoreDelta;
  let playerBScore = game.getPlayerScore(playerIdB) + playerBScoreDelta;
  let isRedScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED)) || judgment.isRedScored;
  let maxGameEndScore = genericFetchUtils.fetchGameEndScore(game);

  logger.debug({key: gameKey}, 'Scores at the end of current turn are for ',
      playerIdA, playerAScore, 'and', playerIdB, playerBScore, 'is red scored',
      isRedScored);
  // game end condition, red should have been scored and all coins of either of the player should be potted
  return (playerAScore >= maxGameEndScore || playerBScore >= maxGameEndScore) &&
      isRedScored;
};

const endGame = function(game, winType) {
  let allPlayerIds = game.getAllPlayerIds();
  if (allPlayerIds.length !== 2) throw new Error(
      'Number of playerIds available is not 2. ' + allPlayerIds.length);
  let playerIdA = allPlayerIds[0];
  let playerIdB = allPlayerIds[1];

  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let playerAScore = game.getPlayerScore(playerIdA);
  let playerBScore = game.getPlayerScore(playerIdB);
  let isRedScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED));
  let redScoredBy;

  // get who scored red
  if (isRedScored)
    redScoredBy = oddUtils.parseBoolean(
        game.getPlayerProp(playerIdA, ALL_PROP_KEYS.DID_SCORED_RED)) ?
        playerIdA :
        playerIdB;
  game.endGame({
    playerAScore: playerAScore,
    playerBScore: playerBScore,
    winType: winType,
    redScoredBy: redScoredBy,
  });
};

const fetchWinnersList = function(game) {
  const gameKey = game.getId();
  let allPlayerIds = game.getAllPlayerIds();
  if (allPlayerIds.length !== 2) throw new Error(
      'Number of playerIds available is not 2. ' + allPlayerIds.length);
  let playerIdA = allPlayerIds[0];
  let playerIdB = allPlayerIds[1];

  let isGameEnded = shouldGameBeEnded(game,
      {isRedScored: false, selfScore: 0, oppScore: 0}, 1);
  if (!isGameEnded)
    logger.warn({key: gameKey},
        'Fetching winners list prematruely before game is ended');
  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let playerAScore = game.getPlayerScore(playerIdA);
  let playerBScore = game.getPlayerScore(playerIdB);
  let isRedScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED));
  let redScoredBy;

  // get who scored red
  if (isRedScored)
    redScoredBy = oddUtils.parseBoolean(
        game.getPlayerProp(playerIdA, ALL_PROP_KEYS.DID_SCORED_RED)) ?
        playerIdA :
        playerIdB;

  let winnerId, loserId;
  if (playerAScore > playerBScore) {
    winnerId = playerIdA;
    loserId = playerIdB;
  } else if (playerAScore < playerBScore) {
    winnerId = playerIdB;
    loserId = playerIdA;
  } else if (isRedScored) {
    winnerId = redScoredBy === playerIdA ? playerIdA : playerIdB;
    loserId = winnerId === playerIdA ? playerIdB : playerIdA;
  }
  return [
    {playerId: winnerId ? winnerId : playerIdA, winType: winnerId ? 1 : 0},
    {playerId: loserId ? loserId : playerIdB, winType: loserId ? 2 : 0},
  ];
};

const onRestoreFromSavepoint = function(game, oldGame) {
  const gameKey = game.getId();
  let duration = genericFetchUtils.fetchMeta(game).turnDuration || 25000;
  let turnExpiry = Date.now() + duration;
  logger.trace({key: gameKey}, 'Setting turn expiry as', turnExpiry,
      'current time is ',
      Date.now());
  game.setGameProp('turnExpiry', turnExpiry);

  for (let p of game.getAllPlayerIds()) {
    game.setPlayerProp(p, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        oldGame.getPlayerProp(p, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT));
    grantGraceTimeForTurn(game, p);
  }

  if (oddUtils.parseBoolean(game.getGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT))) {
    let stkr = genericFetchUtils.fetchStriker(game);
    game.addTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER,
        new TurnAndReconcileStateHolder());
    return {forceIds: {[stkr.id]: true}};
  }
};

const strikerScrutinize = function(game, gameObjects, cb) {
  const gameKey = game.getId();
  let stkr = genericFetchUtils.fetchStriker(game);
  game.forceUpdateSyncVars(false, ...gameObjects);
  for (let obj of gameObjects) if (stkr.id === obj.id && obj.force) {
    game.setGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT, true);
    game.forceUpdateForceGameObjects(false, obj);
    game.issueSavepoint('__force_applied', err => {
      cb(err);
      if (err) return logger.error({err: err, key: gameKey},
          'Error in issuing savepoint');
      logger.debug({key: gameKey}, 'striker hit save point issued');
      game.syncGameStateServVars();
    });
    game.addTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER,
        new TurnAndReconcileStateHolder());
    return;
  }
  cb(null);
};

const abandonMatch = function(game) {
  const gameKey = game.getId();
  logger.info({key: gameKey},
      'Found issues in game state, abandoning the match');
  game.abandonGame('Carrom Board Mismatch');
};

const revertToPreviousSavePoint = function(game) {
  const gameKey = game.getId();
  if (game.getState() >=30)
    return logger.warn({key: gameKey}, 'Game has ended, nothing to do');
  let savepointName = '__turn_start';
  if (oddUtils.parseBoolean(game.getGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT)))
    savepointName = '__force_applied';
  game.restoreFromSavepoint(savepointName, (err, res) => {
    if (err) {
      logger.error({key: gameKey},
          'Error reverting to previous save point. Abandoning match');
      return abandonMatch(res);
    }
  });
};

const onServerRestart = function(game) {
  const gameKey = game.getId();
  if (game.getState() >= 30)
    return logger.debug({key: gameKey}, 'Game has ended, nothing to do');
  game.setGameProp(ALL_PROP_KEYS.SHOULD_SEND_GAME_STATE, false);
  game.restoreFromSavepoint('__turn_start', (err, res) => {
    if (err)
      return logger.error({key: gameKey},
          'Error reverting to prev savepoint, post server restart',
          'continuing game without restore', err);
    logger.info({key: gameKey},
        'Successfully reverted to previous state, post server restart');
    res.setGameProp(ALL_PROP_KEYS.SHOULD_SEND_GAME_STATE, true);
  });
};

module.exports = {
  startGame,
  playerJoined,
  endGame,
  fetchWinnersList,
  shouldGameBeEnded,
  nextTurn,
  continueTurn,
  onRestoreFromSavepoint,
  strikerScrutinize,
  onServerRestart,
  abandonMatch,
  revertToPreviousSavePoint,
  grantGraceTimeForTurn,
};

==> playship-carroms-1.1.9.tgz: package/src/logic/wheeloflife.js <==
const oddUtils = require('../utils/oddUtils');
const logger = oddUtils.getLogger();
const turnFillUtils = require('../utils/fillAndFetchUtils').turnFillUtils;
const genericFetchUtils = require(
    '../utils/fillAndFetchUtils').genericFetchUtils;
const TurnAndReconcileStateHolder = require('../utils/turnAndReconHolder');

const ALL_PROP_KEYS = oddUtils.getAllPropKeys();
const ALL_PROP_VALUES = oddUtils.getALLPropValues();
const ALL_TRANSIENT_KEYS = oddUtils.getAllTransientKeys();

const nextTurn = function(game, isFoul) {
  turnInit(game, parseInt(game.getCurrentTurnIndex()) + 1);
  game.nextTurn(isFoul);
  // issue save point here
  issueTurnStartSavePoint(game);
};

const continueTurn = function(game, isFoul) {
  tackInit(game, parseInt(game.getCurrentTurnIndex()));
  game.continueTurn(isFoul);
  // issue save point here
  issueTurnStartSavePoint(game);
};

function issueTurnStartSavePoint(game) {
  const gameKey = game.getId();
  game.issueSavepoint('__turn_start', err => {
    if (err) return logger.error({err: err, key: gameKey},
        'Error in issuing savepoint');
    logger.debug({key: gameKey}, 'pre-turn save point issued');
  });
}

function turnInit(game, turnIndex) {
  game.setGameProp(ALL_PROP_KEYS.POTT_TACK, 0);
  // initialize empty turn cache for fill fetch utils
  turnFillUtils.initEmptyTurnCache(game, turnIndex | 0);
  tackInit(game, turnIndex | 0);
}

function tackInit(game, currentTurnIndex) {
  let currentTack = parseInt(game.getGameProp(ALL_PROP_KEYS.POTT_TACK));
  // initialize empty tack cache
  turnFillUtils.initEmptyTackCache(game, currentTurnIndex,
      currentTack + 1);
  // increment and set current tack
  game.setGameProp(ALL_PROP_KEYS.POTT_TACK, currentTack + 1);
  let striker = genericFetchUtils.fetchStriker(game);
  striker.setProp(ALL_PROP_KEYS.IS_POTTED, false);

  game.setGameProp(ALL_PROP_KEYS.RECONCILE_STATUS, null);
  game.removeTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER);
  game.removeTransient(ALL_TRANSIENT_KEYS.FORCE_REPLAY_COUNT);

  for (let p of game.getAllPlayerIds()) {
    game.setPlayerProp(p, ALL_PROP_KEYS.GRACE_TIME_START, null);
    game.setPlayerProp(p, ALL_PROP_KEYS.IS_FORCE_TO_BE_REPLAYED, false);
  }

  game.setGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT, false);
}

const grantGraceTimeForTurn = function(game, playerId) {
  let prevGraceTimeLeftForTurn = parseInt(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.GRACE_TIME_LEFT_FOR_TURN));
  let totalGraceTimeLeft = parseInt(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT));

  // bucketing the grace time, for each turn
  let graceTimeToGrantForCurTurn = Math.max(
      ALL_PROP_VALUES.MIN_GRACE_TIME_TO_GRANT,
      Math.min(totalGraceTimeLeft, ALL_PROP_VALUES.GRACE_TIME_BUCKET_FOR_TURN));

  game.setPlayerProp(playerId, ALL_PROP_KEYS.GRACE_TIME_LEFT_FOR_TURN,
      graceTimeToGrantForCurTurn);
  return prevGraceTimeLeftForTurn !== graceTimeToGrantForCurTurn;
};

const startGame = function(game, cb) {
  game.initCollisionMap(cb);
};

const playerJoined = function(game, playerId) {
  const gameKey = game.getId();
  let oppId = genericFetchUtils.fetchOpponentId(game, playerId);
  let playerColor = game.getPlayerProp(playerId, ALL_PROP_KEYS.COLOR);
  let oppColor = game.getPlayerProp(oppId, ALL_PROP_KEYS.COLOR);
  if (!playerColor || !oppColor) {
    // if testing always first player is white, black
    if (process.env.TESTING) {
      playerColor = ALL_PROP_VALUES.COLOR.WHITE;
      oppColor = ALL_PROP_VALUES.COLOR.BLACK;
    } else {
      playerColor = Math.random() * 2 | 0 ?
          ALL_PROP_VALUES.COLOR.WHITE :
          ALL_PROP_VALUES.COLOR.BLACK;
      oppColor = playerColor === ALL_PROP_VALUES.COLOR.WHITE ?
          ALL_PROP_VALUES.COLOR.BLACK :
          ALL_PROP_VALUES.COLOR.WHITE;
    }
    logger.debug({key: gameKey}, 'Setting color', playerColor, 'for', playerId,
        'and', oppColor, 'for', oppId);
    game.setPlayerProp(playerId, ALL_PROP_KEYS.COLOR, playerColor);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.COLOR, oppColor);
    // for 1 player set rotation true
    game.setPlayerProp(playerId, ALL_PROP_KEYS.IS_ROTATION, false);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.IS_ROTATION, true);

    game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_POCKETED_RED, false);
    game.setPlayerProp(playerId, ALL_PROP_KEYS.DID_SCORED_RED, false);

    game.setPlayerProp(oppId, ALL_PROP_KEYS.DID_POCKETED_RED, false);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.DID_SCORED_RED, false);

    game.setPlayerProp(playerId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        ALL_PROP_VALUES.GRACE_TIME_GRANTED);
    game.setPlayerProp(oppId, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        ALL_PROP_VALUES.GRACE_TIME_GRANTED);

    grantGraceTimeForTurn(game, playerId);
    grantGraceTimeForTurn(game, oppId);

    turnInit(game);
    issueTurnStartSavePoint(game);
  }

  // if player has the turn and connected back when force needs to be replayed
  // just revert to previous save point
  let curPlId = game.getCurrentTurnPlayerId();
  if (curPlId === playerId && oddUtils.parseBoolean(
      game.getPlayerProp(playerId, ALL_PROP_KEYS.IS_FORCE_TO_BE_REPLAYED))) {
    logger.info({key: gameKey}, 'Force needs to be replyaed fo the game', game.getTransient(ALL_TRANSIENT_KEYS.FORCE_REPLAY_COUNT));
    // if force needs to be replayed check how many times it was replayed for the turn and forfeit
    if (game.getTransient(ALL_TRANSIENT_KEYS.FORCE_REPLAY_COUNT) > 1)
      return game.forfeitGameForPlayer(playerId, 'Too many force replays');
    revertToPreviousSavePoint(game);
    return game.getState() >= 30; // telling neutrino to not send game state, if game
  }
  // based on should send game, default value is true
  return oddUtils.parseBoolean(
      game.getGameProp(ALL_PROP_KEYS.SHOULD_SEND_GAME_STATE), true);
};

const shouldGameBeEnded = function(game, judgment, playerId) {
  const gameKey = game.getId();
  let allPlayerIds = game.getAllPlayerIds();
  if (allPlayerIds.length !== 2) throw new Error(
      'Number of playerIds available is not 2. ' + allPlayerIds.length);
  let playerIdA = allPlayerIds[0];
  let playerIdB = allPlayerIds[1];

  let playerAScoreDelta, playerBScoreDelta;
  if (playerId === playerIdA) {
    playerAScoreDelta = judgment.selfScore;
    playerBScoreDelta = judgment.oppScore;
  } else {
    playerBScoreDelta = judgment.selfScore;
    playerAScoreDelta = judgment.oppScore;
  }

  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let playerAScore = game.getPlayerScore(playerIdA) + playerAScoreDelta;
  let playerBScore = game.getPlayerScore(playerIdB) + playerBScoreDelta;
  let isRedScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED)) || judgment.isRedScored;
  let maxGameEndScore = genericFetchUtils.fetchGameEndScore(game);

  logger.debug({key: gameKey}, 'Scores at the end of current turn are for ',
      playerIdA, playerAScore, 'and', playerIdB, playerBScore, 'is red scored',
      isRedScored);
  // game end condition, red should have been scored and all coins of either of the player should be potted
  return (playerAScore >= maxGameEndScore || playerBScore >= maxGameEndScore) &&
      isRedScored;
};

const endGame = function(game, winType) {
  let allPlayerIds = game.getAllPlayerIds();
  if (allPlayerIds.length !== 2) throw new Error(
      'Number of playerIds available is not 2. ' + allPlayerIds.length);
  let playerIdA = allPlayerIds[0];
  let playerIdB = allPlayerIds[1];

  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let playerAScore = game.getPlayerScore(playerIdA);
  let playerBScore = game.getPlayerScore(playerIdB);
  let isRedScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED));
  let redScoredBy;

  // get who scored red
  if (isRedScored)
    redScoredBy = oddUtils.parseBoolean(
        game.getPlayerProp(playerIdA, ALL_PROP_KEYS.DID_SCORED_RED)) ?
        playerIdA :
        playerIdB;
  game.endGame({
    playerAScore: playerAScore,
    playerBScore: playerBScore,
    winType: winType,
    redScoredBy: redScoredBy,
  });
};

const fetchWinnersList = function(game) {
  const gameKey = game.getId();
  let allPlayerIds = game.getAllPlayerIds();
  if (allPlayerIds.length !== 2) throw new Error(
      'Number of playerIds available is not 2. ' + allPlayerIds.length);
  let playerIdA = allPlayerIds[0];
  let playerIdB = allPlayerIds[1];

  let isGameEnded = shouldGameBeEnded(game,
      {isRedScored: false, selfScore: 0, oppScore: 0}, 1);
  if (!isGameEnded)
    logger.warn({key: gameKey},
        'Fetching winners list prematruely before game is ended');
  let redCoin = genericFetchUtils.fetchRedCoin(game);
  let playerAScore = game.getPlayerScore(playerIdA);
  let playerBScore = game.getPlayerScore(playerIdB);
  let isRedScored = oddUtils.parseBoolean(
      redCoin.getProp(ALL_PROP_KEYS.IS_SCORED));
  let redScoredBy;

  // get who scored red
  if (isRedScored)
    redScoredBy = oddUtils.parseBoolean(
        game.getPlayerProp(playerIdA, ALL_PROP_KEYS.DID_SCORED_RED)) ?
        playerIdA :
        playerIdB;

  let winnerId, loserId;
  if (playerAScore > playerBScore) {
    winnerId = playerIdA;
    loserId = playerIdB;
  } else if (playerAScore < playerBScore) {
    winnerId = playerIdB;
    loserId = playerIdA;
  } else if (isRedScored) {
    winnerId = redScoredBy === playerIdA ? playerIdA : playerIdB;
    loserId = winnerId === playerIdA ? playerIdB : playerIdA;
  }
  return [
    {playerId: winnerId ? winnerId : playerIdA, winType: winnerId ? 1 : 0},
    {playerId: loserId ? loserId : playerIdB, winType: loserId ? 2 : 0},
  ];
};

const onRestoreFromSavepoint = function(game, oldGame) {
  const gameKey = game.getId();
  let duration = genericFetchUtils.fetchMeta(game).turnDuration || 25000;
  let turnExpiry = Date.now() + duration;
  logger.trace({key: gameKey}, 'Setting turn expiry as', turnExpiry,
      'current time is ',
      Date.now());
  game.setGameProp('turnExpiry', turnExpiry);

  for (let p of game.getAllPlayerIds()) {
    game.setPlayerProp(p, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT,
        oldGame.getPlayerProp(p, ALL_PROP_KEYS.TOTAL_GRACE_TIME_LEFT));
    grantGraceTimeForTurn(game, p);
  }

  if (oddUtils.parseBoolean(game.getGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT))) {
    let stkr = genericFetchUtils.fetchStriker(game);
    game.addTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER,
        new TurnAndReconcileStateHolder());
    game.addTransient(ALL_TRANSIENT_KEYS.FORCE_REPLAY_COUNT,
        (oldGame.getTransient(ALL_TRANSIENT_KEYS.FORCE_REPLAY_COUNT) || 0) + 1);
    return {forceIds: {[stkr.id]: true}};
  }
};

const strikerScrutinize = function(game, gameObjects, cb) {
  const gameKey = game.getId();
  let stkr = genericFetchUtils.fetchStriker(game);
  game.forceUpdateSyncVars(false, ...gameObjects);
  for (let obj of gameObjects) if (stkr.id === obj.id && obj.force) {
    game.setGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT, true);
    game.forceUpdateForceGameObjects(false, obj);
    game.issueSavepoint('__force_applied', err => {
      cb(err);
      if (err) return logger.error({err: err, key: gameKey},
          'Error in issuing savepoint');
      logger.debug({key: gameKey}, 'striker hit save point issued');
      game.syncGameStateServVars();
    });
    game.addTransient(ALL_TRANSIENT_KEYS.TURN_RECONCILE_HOLDER,
        new TurnAndReconcileStateHolder());
    return;
  }
  cb(null);
};

const abandonMatch = function(game) {
  const gameKey = game.getId();
  logger.info({key: gameKey},
      'Found issues in game state, abandoning the match');
  game.abandonGame('Carrom Board Mismatch');
};

const revertToPreviousSavePoint = function(game) {
  const gameKey = game.getId();
  if (game.getState() >= 30)
    return logger.warn({key: gameKey}, 'Game has ended, nothing to do');
  let savepointName = '__turn_start';
  let isStrikerHit = oddUtils.parseBoolean(
      game.getGameProp(ALL_PROP_KEYS.IS_STRIKER_HIT));
  if (isStrikerHit) savepointName = '__force_applied';
  game.restoreFromSavepoint(savepointName, (err, res) => {
    if (err) {
      logger.error({key: gameKey},
          'Error reverting to previous save point. Abandoning match');
      return abandonMatch(res);
    }
  });
};

const onServerRestart = function(game) {
  const gameKey = game.getId();
  if (game.getState() >= 30)
    return logger.debug({key: gameKey}, 'Game has ended, nothing to do');
  game.setGameProp(ALL_PROP_KEYS.SHOULD_SEND_GAME_STATE, false);
  game.restoreFromSavepoint('__turn_start', (err, res) => {
    if (err)
      return logger.error({key: gameKey},
          'Error reverting to prev savepoint, post server restart',
          'continuing game without restore', err);
    logger.info({key: gameKey},
        'Successfully reverted to previous state, post server restart');
    res.setGameProp(ALL_PROP_KEYS.SHOULD_SEND_GAME_STATE, true);
  });
};

module.exports = {
  startGame,
  playerJoined,
  endGame,
  fetchWinnersList,
  shouldGameBeEnded,
  nextTurn,
  continueTurn,
  onRestoreFromSavepoint,
  strikerScrutinize,
  onServerRestart,
  abandonMatch,
  revertToPreviousSavePoint,
  grantGraceTimeForTurn,
};
